/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Parts of this code are taken from other
    filesystem drivers also with GPL licenses:
    swapfs, minix, romfs, affs.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/locks.h>

#include <asm/uaccess.h>

/* some random number */
#define TSTORE_MAGIC	0x74617273

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/locks.h>
#include <linux/init.h>
#include <linux/smp_lock.h>

#include <asm/uaccess.h>

//#define TSTORE_DBG(x) printk("%s %s\n",x,__FUNCTION__)
#define TSTORE_DBG(x) do { } while(0)
#define ENTER() TSTORE_DBG("ENTER")

/* Do this like affs, I also need to know */
extern int *blk_size[];

static struct super_operations tstore_ops;

uint64_t get_nbyteint(uint8_t *p, int n)
{
  uint64_t r=0;
  int i;
  for (i=0;i<n;++i)
    r=(r<<8)+p[i];
  return r;
}

void put_nbyteint(uint8_t *p, int n, uint64_t v)
{
  int i;
  for (i=0;i<n;++i)
    p[i]=v>>(((n-1)-i)*8);
}

#define PUT_BNUM(_p,_v) put_nbyteint(_p,5,_v)
#define GET_BNUM(_p)    get_nbyteint(_p,5)
#define PUT_SIZE(_p,_v) put_nbyteint(_p,7,_v)
#define GET_SIZE(_p)    get_nbyteint(_p,7)

int filelevels(uint64_t size)
{
  int levels=0;
  uint64_t maxsize=512;
  while(size>maxsize) {
    levels++;
    maxsize*=102;
  }
  return levels;
}

static struct super_block *
tstore_read_super(struct super_block *s, void *data, int silent)
{
	struct buffer_head *bh;
	kdev_t dev = s->s_dev;
	struct tstore_super_block *rsb;
	int blocks;

	ENTER();

	/* I would parse the options here, but there are none.. :) */

	set_blocksize(dev, 512);
	s->s_blocksize = 512;
	s->s_blocksize_bits = 9;
	s->u.generic_sbp = (void *) 0;
	s->s_maxbytes = 0xFFFFFFFF;

	bh = sb_bread(s, 0);
	if (!bh) {
		/* XXX merge with other printk? */
                printk ("TSTORE: unable to read superblock\n");
		goto outnobh;
	}

        blocks = blk_size[MAJOR(dev)] ? blk_size[MAJOR(dev)][MINOR(dev)] : 0;
        if (!blocks) {
                printk(KERN_ERR "TSTORE: Could not determine device size\n");
                goto out;
        }

	rsb = (struct tstore_super_block *)bh->b_data;
	if (memcmp(bh->b_data,"TSTORE2",8)&&memcmp(bh->b_data,"TSTORE3",8)) {
		if (!silent)
			printk ("VFS: Can't find a tstore filesystem on dev "
				"%s.\n", kdevname(dev));
		goto out;
	}

	brelse(bh);

	/* Find the start of the fs */

	blocks <<= 1; /* FIXME: Is this always correct? */

	bh = sb_bread(s, blocks);
	if (bh) {
		uint8_t *buf=bh->b_data;
		if (buf[17]&&!buf[511]) {
			int i;
			for (i=17;buf[i]&96;++i);
			for (;(i<512)&&!buf[i];++i);
			if (i==512) ++blocks;
		}
		brelse(bh);
	}

	/* Try to read the root */
	/*
	bh = sb_bread(s, blocks-1);
	if (bh) {
		printk("%s\n",bh->b_data+17);
		brelse(bh);
	}
	*/

	s->s_magic = TSTORE_MAGIC;
	s->u.romfs_sb.s_maxsize = blocks;

	s->s_flags |= MS_RDONLY;

	s->s_op	= &tstore_ops;
	s->s_root = d_alloc_root(iget(s, blocks));

	if (!s->s_root)
		goto outnobh;

	/* Ehrhm; sorry.. :)  And thanks to Hans-Joachim Widmaier  :) */
	if (0) {
out:
		brelse(bh);
outnobh:
		s = NULL;
	}

	return s;
}

/* That's simple too. */

static int
tstore_statfs(struct super_block *sb, struct statfs *buf)
{
	ENTER();
	buf->f_type = TSTORE_MAGIC;
	buf->f_bsize = 512;
	buf->f_bfree = buf->f_bavail = buf->f_ffree;
	buf->f_blocks =  sb->u.romfs_sb.s_maxsize;
	buf->f_namelen = 494;
	return 0;
}

static long tstore_block_number(struct super_block *sb,
				long ino,
				long block)
{
	int levels;
	struct buffer_head *bh;
	uint8_t *buf;
	uint64_t block_number;
	
	bh=sb_bread(sb,ino);
	if (!bh) {
		printk("What now?\n");
		return -EIO;
	}

	buf=bh->b_data;

	levels=filelevels(GET_SIZE(buf+10));
	block_number=GET_BNUM(buf+5);

	brelse(bh);


	while ((--levels)>=0) {
		int idx=block,x;
		for (x=0;x<levels;++x) idx/=102;
		idx%=102;

		if (block_number & (3ULL<<38)) {
			block_number = (block_number - (1ULL<<38) + idx);
		} else if (block_number) {
			bh=sb_bread(sb,block_number);
			if (!bh) {
				printk("What now?\n");
				return -EIO;
			}
			block_number=GET_BNUM(bh->b_data+5*idx);
			brelse(bh);
		}
	}

	if (block_number & (3ULL<<38)) {
		printk("TSTORE: Final sequential %llx\n",block_number);
		return -EIO;
	}

	return block_number;
}

static int tstore_get_block(struct inode *inode, long block,
			    struct buffer_head *bh_result, int create)
{
	long trans_block;
	/* create should never be set. And if it
	 * happens anyway I'd rather crash right
	 * here than corrupt my storage.
	 */
	BUG_ON(create);
	if (create) panic("This is not possible\n");

	trans_block=tstore_block_number(inode->i_sb,inode->i_ino,block);
	if (trans_block<1) return trans_block;

	bh_result->b_dev = inode->i_dev;
        bh_result->b_blocknr = trans_block;
        bh_result->b_state |= (1UL << BH_Mapped);
        return 0;
}

static int tstore_readpage(struct file *file, struct page *page)
{
  return block_read_full_page(page,tstore_get_block);
}
static int tstore_bmap(struct address_space *mapping, long block)
{
  return generic_block_bmap(mapping,block,tstore_get_block);
}


/*
 * Lookup the data. This needs to be replaced.
 */
static struct dentry * tstore_lookup(struct inode *dir, struct dentry *dentry)
{
	int i;
	struct super_block *sb=dir->i_sb;
	for (i=sb->u.romfs_sb.s_maxsize-1;i;) {
		struct buffer_head *bh;
		uint8_t *buf;

		bh=sb_bread(sb, i);
		if (!bh) return ERR_PTR(-EIO);

		buf=bh->b_data;
		if (!strcmp(dentry->d_name.name,buf+17)) {
			struct inode *inode;
			brelse(bh);
			inode=iget(sb,i);
			if (!inode)
				return ERR_PTR(-EACCES);
			d_add(dentry,inode);
			return NULL;
		}
		i=GET_BNUM(buf);
		brelse(bh);
	}
        d_add(dentry, NULL);
        return NULL;
}

static int
tstore_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	struct dentry *dentry = filp->f_dentry;
	struct super_block *sb = dentry->d_inode->i_sb;
	ino_t ino;
	ENTER();

	switch (filp->f_pos) {
		case 0:
			ino = dentry->d_inode->i_ino;
			if (filldir(dirent, ".", 1, filp->f_pos, ino, DT_DIR) < 0)
				break;
			filp->f_pos=1;
			/* fallthrough */
		case 1:
			spin_lock(&dcache_lock);
			ino = dentry->d_parent->d_inode->i_ino;
			spin_unlock(&dcache_lock);
			if (filldir(dirent, "..", 2, filp->f_pos, ino, DT_DIR) < 0)
				break;
			filp->f_pos=sb->u.romfs_sb.s_maxsize+1;
			/* fallthrough */
		default:
			while(filp->f_pos > 2) {
				struct buffer_head *bh;
				uint8_t *buf;

				bh = sb_bread(sb,filp->f_pos-2);
				if (!bh) return -EIO;

				buf=bh->b_data;

				//printk(">>>%d\n",filp->f_pos);

				if (filldir(dirent, buf+17, strlen(buf+17), filp->f_pos, filp->f_pos-2, DT_REG)) {
					brelse(bh);
					return 0;
				}

				filp->f_pos=GET_BNUM(buf)+2;

				brelse(bh);
			}
	}
	return 0;
}

/* Mapping from our types to the kernel */

static struct address_space_operations tstore_aops = {
	readpage:	tstore_readpage,
	bmap:		tstore_bmap,
};

static struct file_operations tstore_dir_operations = {
	read:		generic_read_dir,
	readdir:	tstore_readdir,
};

static struct inode_operations tstore_dir_inode_operations = {
	lookup:		tstore_lookup,
};

static void tstore_read_inode(struct inode *inode)
{
	int ino=inode->i_ino;
	ENTER();

	inode->i_uid = 0;
	inode->i_gid = 0;
	inode->i_nlink = 1;
	inode->i_size = 0;
	inode->i_blocks = inode->i_blksize = 0;
	inode->i_mode = S_IRUGO|S_IXUGO;
	inode->i_mapping->a_ops = &tstore_aops;
	inode->i_data.a_ops = &tstore_aops;

        inode->i_mtime = inode->i_atime = inode->i_ctime = 0;

	if (ino < inode->i_sb->u.romfs_sb.s_maxsize) {
		struct buffer_head *bh;
		uint8_t *buf;
		/* file */

		bh=sb_bread(inode->i_sb,ino);
		if (!bh) {
			printk("Now what?\n");
		} else {
			buf=bh->b_data;
			inode->i_mode |= S_IFREG;
			inode->i_fop = &generic_ro_fops;
			inode->i_size = GET_SIZE(buf+10);
			//printk("ino=%d max=%d\n",ino,inode->i_sb->u.romfs_sb.s_maxsize);
		}
	} else {
		/* root directory */
		inode->i_mode |= S_IFDIR;
		inode->i_size = 42;
		inode->i_nlink = 2;
		inode->i_op = &tstore_dir_inode_operations;
		inode->i_fop = &tstore_dir_operations;
		//printk("%08X\n",inode->i_mode);
	}
}


static struct super_operations tstore_ops = {
	read_inode:	tstore_read_inode,
	statfs:		tstore_statfs,
};

static DECLARE_FSTYPE_DEV(tstore_fs_type, "tstore", tstore_read_super);

static int __init init_tstore_fs(void)
{
	return register_filesystem(&tstore_fs_type);
}

static void __exit exit_tstore_fs(void)
{
	unregister_filesystem(&tstore_fs_type);
}

/* Yes, works even as a module... :) */

EXPORT_NO_SYMBOLS;

module_init(init_tstore_fs)
module_exit(exit_tstore_fs)
MODULE_LICENSE("GPL");
